defmodule Chronoscope.NTS.Behaviour do
  @callback(key_establishment(host :: String.t(), port :: integer()) :: {:ok, Map.t()}, {:error, any()})
end

defmodule Chronoscope.NTS do
  @behaviour Chronoscope.NTS.Behaviour

  require Logger

  alias Chronoscope.NTS.KeyEstablishment

  @timeout_in_milliseconds 3000

  def key_establishment(host, port) do
    case ssl_connect(host, port) do
      {:ok, socket} -> perform_key_establishment(socket)
      {:error, {:tls_alert, {:handshake_failure, error}}} -> {:error, to_string(error)}
      {:error, :timeout} -> {:error, :timeout}
      {:error, error} -> {:error, inspect(error)}
      error -> {:error, inspect(error)}
    end
  end

  defp ssl_connect(host, port) do
    :ssl.connect(host, port, tls_options(host), @timeout_in_milliseconds)
  end

  defp tls_options(host) do
    host
    |> :tls_certificate_check.options()
    |> Keyword.put(:alpn_advertised_protocols, ["ntske/1"])
    |> Keyword.put(:verify_fun, {&verify_fun/3, [check_hostname: host]})
  end

  defp verify_fun(cert, :valid_peer = event, intial_user_state) do
    {:Validity, {:utcTime, _from}, {:utcTime, _to}} = X509.Certificate.validity(cert)
    # parse datetime from messy erlang version and store in nts client genserver

    :ssl_verify_hostname.verify_fun(cert, event, intial_user_state)
  end

  defp verify_fun(cert, event, initial_user_state) do
    :ssl_verify_hostname.verify_fun(cert, event, initial_user_state)
  end

  defp perform_key_establishment(socket) do
    :ok = :ssl.send(socket, KeyEstablishment.request())

    receive do
      {:ssl, _socket, response} ->
        :ssl.close(socket)
        KeyEstablishment.parse_response(response)

      msg ->
        :ssl.close(socket)
        Logger.error("received unexpected message: #{inspect(msg)}")
        {:error, :no_response}
    after
      @timeout_in_milliseconds ->
        :ssl.close(socket)
        Logger.error("timed out waiting for response")
        {:error, :timeout}
    end
  end
end
